/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

/* 
 * File:   Background.cpp
 * Author: tg
 * 
 * Created on April 26, 2016, 5:39 PM
 */

#include <sstream>
#include <iostream>
#include <cmath>
#include <limits>
#include <algorithm>
#include <iomanip>

#include "Background.h"
#include "io.h"
#include "myExceptions.h"

Background::Background(const std::string& inputdir, const std::string& outbase, unsigned short from, unsigned short to, float bgsigmalevel):
inputdir_(inputdir), outbase_(outbase), from_(from), to_ (to), bgsigmalevel_ (bgsigmalevel) {
    
    
    // read data (again) to make a copy
    size_t peak;
    for (int i = from; i < to; ++i) {
        std::ostringstream num;
        size_t peak;
        num << i;
        std::string input (inputdir_ + infile (i));
        const std::vector<unsigned short> rawdata (IO::in::raw(input.c_str(), width_, height_, peak, false));
        Image<unsigned short> myimage (rawdata, width_, height_);
        imagestack_.push_back(myimage);
    }
    
    if (imagestack_.empty()) {
        std::cout << "# Error: now images in stack\n";
        throw myExcepts::FileMissing ("Emtpy image stack");
    }
    
    stacknoise_ = std::vector<float> (width_*height_, 0.0);
    peaks3D_ = std::vector<short> (imagestack_.size()*width_*height_, 0);

    neighbour_indices_ [0]  = -1;
    neighbour_indices_ [1]  = +1;
    neighbour_indices_ [2]  = -width_ - 1;
    neighbour_indices_ [3]  = -width_;
    neighbour_indices_ [4]  = -width_ + 1;
    neighbour_indices_ [5]  =  width_ - 1;
    neighbour_indices_ [6]  =  width_;
    neighbour_indices_ [7]  =  width_ + 1;
    neighbour_indices_ [8]  = -(width_ * height_) - 1;
    neighbour_indices_ [9]  = -(width_ * height_);
    neighbour_indices_ [10] = -(width_ * height_) + 1;
    neighbour_indices_ [11] = -(width_ * height_) - width_ - 1;
    neighbour_indices_ [12] = -(width_ * height_) - width_;
    neighbour_indices_ [13] = -(width_ * height_) - width_ + 1;
    neighbour_indices_ [14] = -(width_ * height_) + width_ - 1;
    neighbour_indices_ [15] = -(width_ * height_) + width_;
    neighbour_indices_ [16] = -(width_ * height_) + width_ + 1;
    neighbour_indices_ [17] =  (width_ * height_) - 1;
    neighbour_indices_ [18] =  (width_ * height_);
    neighbour_indices_ [19] =  (width_ * height_) + 1;
    neighbour_indices_ [20] =  (width_ * height_) - width_ - 1;
    neighbour_indices_ [21] =  (width_ * height_) - width_;
    neighbour_indices_ [22] =  (width_ * height_) - width_ + 1;
    neighbour_indices_ [23] =  (width_ * height_) + width_ - 1;
    neighbour_indices_ [24] =  (width_ * height_) + width_;
    neighbour_indices_ [25] =  (width_ * height_) + width_ + 1;
}

Background::~Background() {
}

void Background::noise() {
    std::vector <Image<unsigned short>>::const_iterator it;
    
    std::vector<unsigned int> counts (width_*height_, 0);
    
    for (it = imagestack_.begin(); it != imagestack_.end(); ++it) {
        for (int h = 0; h < height_; ++h) {
            size_t idx (h * width_);
            for (int w = 0; w < width_; ++w, ++idx) {
                if (it->masked(idx)) continue;
                const unsigned short val (it->data(idx));
                stacknoise_[idx] += val;
                ++ (counts[idx]);
            }
        }
    }
    std::vector<float>::iterator itn;
    for (itn = stacknoise_.begin(); itn != stacknoise_.end(); ++itn) {
        const size_t idx (itn - stacknoise_.begin());
        *itn /= counts[idx];
    }
}
/**
 * For each pixel in stack, set to zero if < mean + k*sqrt(mean)
 * @param ksigma
 */
void Background::denoise(float ksigma) {

    std::cout << "---> Calculating noise along pixels\n";
    noise();
    
    std::vector <char> mask (peaks3D_.size(), 0);
    std::cout << "---> zeroing all pixels below mu+ " << ksigma << " sigma\n";
    for (int h = 0; h < height_; ++h) {
        size_t idx(h * width_);
        for (int w = 0; w < width_; ++w, ++idx) {
            unsigned int cutoff(stacknoise_[idx] + ksigma * std::sqrt(stacknoise_[idx]));
            std::vector <Image<unsigned short>>::iterator it;
            for (it = imagestack_.begin(); it != imagestack_.end(); ++it) {
                if (it->data(idx) < cutoff) {
                    it->data(idx) = 0;
                    size_t maskidx (w + width_*(h+height_*(it-imagestack_.begin())));
                    mask [maskidx] = 1;
                }
            }
        }
    }
    std::cout << "---> Connecting contiguous regions\n";
    unsigned short label (0);
    std::vector<char>::const_iterator it;
    for (it = mask.begin(); it != mask.end(); ++it) {
        if (*it == 0) {
            size_t idx = it - mask.begin();
            ++label;
            searchNeighbours(idx, label);
            if (label >= std::numeric_limits<unsigned short>::max()) {
                std::cout << "---> Search reached max. number of peaks ("
                        << label << ")\n";
                break;
            }
        }
    }
    std::cout << "Number of contiguous regions: " << label << "\n";
    
    
//    std::cout << "---> Collection of contiguous non-zero pixels\n";
//    despeckle();
//    despeckle();
}

/**
 * Noise reduction by convolution with the Sobel operator, see Wikipedia
 * https://en.wikipedia.org/wiki/Sobel_operator
 */
void Background::sobel() {
    std::vector <Image<unsigned short>>::iterator it;
    const short sobelx [] = { 1, 0, -1, 2, 0, -2, 1, 0, -1};
    const short sobely [] = { 1, 2, 1 , 0, 0, 0, -1, -2, -1};
    
    for (it = imagestack_.begin(); it != imagestack_.end(); ++it) {
        std::vector <unsigned short> edges (width_*height_, 0);
        for (short h = 1; h < height_ -1; ++h) {
            for (short w = 1; w < width_-1; ++w) {
                int gx(0);
                int gy(0);
                
                for (short i = 0; i < 3; ++i) {
                    for (short j = 0;  j < 3; ++j) {
                        size_t idx ( (w + i -1) + (h +j -1)* width_);
                        gx += it->data(idx) * sobelx [ i + j*3];
                        gy += it->data(idx) * sobely [ i + j*3];
                    }
                }
                edges [w + h * width_] = short(std::sqrt(gx*gx + gy*gy));
            }
        }
        *it = Image<unsigned short> (edges, width_, height_);
    }
}

/**
 * remove pixels w/o neighbouring pixels. Only check orthogonal neighbours per 
 * frame and only the same pixel on the next frame
 */
void Background::despeckle() {
    std::vector <Image<unsigned short>>::iterator it;

    // create 3D volume for peaks
    for (it = imagestack_.begin(); it != imagestack_.end() - 1; ++it) {
        for (short h = 1; h < height_ - 1; ++h) {
            for (short w = 1; w < width_ - 1; ++w) {
                size_t idx(w + h * width_);
                if (it->data(idx) == 0) continue;

                if (it->data(idx - 1) == 0 &&
                        it->data(idx + 1) == 0 &&
                        it->data(idx - width_) == 0 &&
                        it->data(idx + width_) == 0 &&
                        (it + 1)->data(idx) == 0) {
                    it->data(idx) = 0;
                    // next w already checked
                    //++w;
                }
            }
        }
    }
}

void Background::searchNeighbours(const size_t index, const short label) {
    std::vector<int>::const_iterator it;
    peaks3D_[index] = label;

    peakArea_.imagenumber = index / (width_*height_);
    size_t dataidx = index % (width_ * height_);
    peakArea_.h = dataidx / width_;
    peakArea_.w = index % width_;
    peakArea_.val = imagestack_[peakArea_.imagenumber].data(dataidx);
    labelledPeaks_[label].push_back(peakArea_);

    for (int i = 0; i < 26; ++i) {
        int nextidx = index + neighbour_indices_[i];
        if (nextidx < 0 || nextidx > peaks3D_.size()) continue;
        int nextimage = nextidx / (width_*height_);
        size_t nextdataidx = nextidx % (width_ * height_);;
        if (peaks3D_[nextidx] == 0 && imagestack_[nextimage].data(nextdataidx) != 0) {
            searchNeighbours(nextidx, label);
        }
    }

}


void Background::writeImages() {
    std::vector <Image<unsigned short>>::const_iterator it;
    
    for (it = imagestack_.begin(); it != imagestack_.end(); ++it) {
        Image <int> image (it->data(), it->width(), it->height());
        
        for (size_t h = 0; h < image.height(); ++h) {
            size_t idx(0 + h * image.width());
            image.data(idx) = -1;
            idx = 255 + h * image.width();
            image.data(idx) = -1;
            idx = 256 + h * image.width();
            image.data(idx) = -1;
            idx = 511 + h * image.width();
            image.data(idx) = -1;
        }
        for (size_t w = 0; w < image.height(); ++w) {
            size_t idx(w + 0 * image.width());
            image.data(idx) = -1;
            idx = w + 255 * image.width();
            image.data(idx) = -1;
            idx = w + 256 * image.width();
            image.data(idx) = -1;
            idx = w + 511 * image.width();
            image.data(idx) = -1;
        }

        size_t idx (it - imagestack_.begin() + from_);
        std::string output (outfile (outbase_, idx, to_));

        IO::out::cbf(output.c_str(), image);
    }

}

/**
 * Create input filename of the form frame_value_i.bin
 * @param i
 * @return 
 */
std::string Background::infile(unsigned short i) {
    std::ostringstream outp;
    outp << "frame_value_" << i << ".bin";
    return outp.str();
}

std::string Background::outfile(const std::string& base, unsigned short i, unsigned short spacer) {
    unsigned short digits (1);
    if (spacer == 0) {
        throw myExcepts::Usage ("Spacer must be > 0");
    }
    
    if (i > spacer) {
        digits = short(log10(i)+1);
    }
    else {
        digits = short(log10(spacer)+1);
    }
    std::ostringstream outp;
    outp << base;
    if (*(base.rbegin()) != '_') {
        outp << "_";
    }
    outp << "denoised_" 
            << std::setw(digits) << std::setfill('0') << i << ".cbf";
    return outp.str();
}