/* 
 * File:   Image.h
 * Author: tg
 *
 * Created on December 11, 2013, 11:03 PM
 */

#ifndef IMAGE_H
#define	IMAGE_H

#include <vector>


/**
 * \class Image
 * \brief Two-dimensional  array with integer data points
 * Provides function for peak search to return an array of XY with
 * peak height and a sigma-estimate.
 * I assume that the data of the image and its mask fits into computer memory
 */
template <typename T> class Image {
private:
    std::vector<T> data_;
    int width_;
    int height_;
    std::vector<bool> mask_;
    long int peak_;
    
    //! summed simple gradient
    void gradient (const int cx, const int cy, const char R, int& gx, int& gy) const;
    //! optimal line step within maxsteps
    void linestep (int& x0, int& y0, const char R, const int gx, const int gy, int& pixelsum,
            const int maxsteps) const;
    //! sum of pixel within square
    int sumpixels (const int x0, const int y0, const char R) const;
    
public:
    Image(const std::vector<unsigned short>& data, const int& width, 
            const int& height, const long int peak=0, const short crosscale=3);
    ~Image();
    
    //! set entire mask
    void setmask (const std::vector<char>& mask);

    //! set specific pixel (w/h) in mask
    void setmask (int w, int h);
    
    //! return dimensions
    int width() const { return width_; }
    int height() const { return height_; }
    
    std::vector<T> data() const { return data_; }
    
    // return value at (w, h)
    T data(const unsigned int idx) const { return data_[idx]; }
    
    // return reference for setting data point
    T& data (const unsigned int idx) { return data_[idx]; }
    
    // return mask at (w/h)
    bool masked (const unsigned int idx) const { return mask_[idx]; }
    
    //! calculate density centre of images, starting at x0, y0
    void centre (const char radius, int& x0, int& y0) const;
    
    //! add pixel values to an int-image
    void add (Image<int>& image) const;
    
    // return the peaks
    //std::vector<Maximum> peaks(unsigned short peakfactor, bool verbose) const;
};



#endif	/* IMAGE_H */

