/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

/* 
 * File:   io.h
 * Author: tg
 *
 * Created on April 18, 2016, 4:24 PM
 */

#ifndef IO_H
#define IO_H

#include "Image.h"
#include <vector>
#include <tiffio.h>

/**
 * Input and output routines
 */
namespace IO {
    namespace in {
        // timepix binary format: unsigned short
        std::vector<unsigned short> raw (const char* filename, 
                unsigned short& width, unsigned short& height, 
                size_t& peak, bool verbose = false);
        std::vector<unsigned short> tiff (const char* filename, 
                unsigned short& width, unsigned short& height, 
                bool verbose = false);

    }
    
    namespace out {
        // write image values to text file
        void txt (const char* filename, const Image<unsigned short>& image);
        
        // create a cbf file readable by XDS
        void cbf (const char* filename, const Image<int>& image);
    } 
}

namespace Utils {
    // parser for cbfheader for timepix data
    std::string cbfheader_timepix (const Image<int>& image);
    
    //! create a cbfheader from given information
    std::string cbfheader(const std::string detector, const std::string sn, 
        float pixel_x_mu, float pixel_y_mu,
        float silicon_thickness,
        float exposure,
        int cutoff,
        float threshold,
        int nexclpixels,
        const std::string& flatfield,
        float wavelength,
        float detdistance,
        float detvoffset,
        float beamx, float beamy,
        float alpha_start, float delta_alpha);
    
    std::string timestamp();
    
    
}


#endif /* IO_H */

